#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017 Ximin Luo <infinity0@debian.org>
# Copyright © 2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import json

from ..difference import Difference
from ..presenters.json import JSON_FORMAT_MAGIC

from .utils import UnrecognizedFormatError


class JSONReaderV1:
    def load(self, fp, fn):
        # fp should be a str-stream not a bytes-stream. If you need to pass in
        # a bytes-stream, wrap it in codecs.getreader('utf-8')(fp)
        raw = json.load(fp)
        if JSON_FORMAT_MAGIC not in raw or raw[JSON_FORMAT_MAGIC] != 1:
            raise UnrecognizedFormatError(
                "Magic not found in JSON: {}".format(JSON_FORMAT_MAGIC)
            )
        return self.load_rec(raw)

    def load_rec(self, raw):
        source1 = raw["source1"]
        source2 = raw["source2"]
        unified_diff = raw["unified_diff"]
        comments = raw.get("comments", [])
        details = [self.load_rec(child) for child in raw.get("details", [])]

        return Difference(
            source1,
            source2,
            comment=comments,
            details=details,
            unified_diff=unified_diff,
        )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import zlib

from diffoscope.difference import Difference

from .utils.file import File

MAGIC = b"# Sphinx inventory version"


class SphinxInventoryFile(File):
    DESCRIPTION = "Sphinx inventory files"
    FILE_EXTENSION_SUFFIX = {".inv"}

    def compare_details(self, other, source=None):
        return [
            Difference.from_text(
                describe_inventory(self.path),
                describe_inventory(other.path),
                self.path,
                other.path,
                source="Sphinx inventory",
            )
        ]


def describe_inventory(filename):
    head = b""
    tail = b""

    with open(filename, "rb") as f:
        for line in f:
            if line.startswith(b"#"):
                # Save commented lines at top of file
                head += line
            else:
                # ... save the rest for decompression
                tail += line

    result = head + b"\n" if head else b""

    try:
        result += zlib.decompress(tail)
    except zlib.error:
        # Even if the file is labelled "The remainder of this file is
        # compressed using zlib.", it might not actually be. In this case,
        # simply return the original content.
        result += tail

    return result.decode("utf-8", "ignore")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015-2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import sys
import logging

from diffoscope.diff import color_unified_diff
from diffoscope.config import Config

from .utils import (
    Presenter,
    create_limited_print_func,
    PrintLimitReached,
    make_printer,
)

logger = logging.getLogger(__name__)


class TextPresenter(Presenter):
    PREFIX = u"│ "
    RE_PREFIX = re.compile(r"(^|\n)")

    def __init__(self, print_func, color):
        self.print_func = create_limited_print_func(
            print_func, Config().max_text_report_size
        )
        self.color = color

        super().__init__()

    @classmethod
    def run(cls, data, difference, parsed_args):
        with make_printer(data["target"]) as fn:
            color = {
                "auto": fn.output.isatty(),
                "never": False,
                "always": True,
            }[parsed_args.text_color]

            presenter = cls(fn, color)

            try:
                presenter.start(difference)
            except UnicodeEncodeError:
                logger.critical(
                    "Console is unable to print Unicode characters. Set e.g. "
                    "PYTHONIOENCODING=utf-8"
                )
                sys.exit(2)

    def start(self, difference):
        try:
            super().start(difference)
        except PrintLimitReached:
            self.print_func("Max text output size reached.", force=True)

    def visit_difference(self, difference):
        if self.depth == 0:
            self.output("--- {}".format(difference.source1))
            self.output("+++ {}".format(difference.source2))
        elif difference.source1 == difference.source2:
            self.output(u"├── {}".format(difference.source1))
        else:
            self.output(u"│   --- {}".format(difference.source1))
            self.output(u"├── +++ {}".format(difference.source2))

        for x in difference.comments:
            self.output(u"│┄ {}".format(x))

        diff = difference.unified_diff

        if diff:
            self.output(color_unified_diff(diff) if self.color else diff, True)

    def output(self, val, raw=False):
        self.print_func(
            self.indent(val, self.PREFIX * (self.depth + (0 if raw else -1)))
        )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import struct
import logging
import platform
import subprocess

from diffoscope.tools import tool_required
from diffoscope.profiling import profile
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command, our_check_output

HI_MAGIC_32 = struct.pack(">I", 0x1FACE)
HI_MAGIC_64 = struct.pack(">I", 0x1FACE64)

if platform.architecture()[0] == "32bit":
    HI_MAGIC = HI_MAGIC_32
else:
    HI_MAGIC = HI_MAGIC_64

logger = logging.getLogger(__name__)


class ShowIface(Command):
    @tool_required("ghc")
    def cmdline(self):
        return ["ghc", "--show-iface", self.path]


class HiFile(File):
    """
    Here is how an example .hi file starts:

    % hexdump -C tests/data/test1.hi | head -n 1
    00000000  01 fa ce 64 00 00 00 00  00 00 00 00 04 00 00 00  |...d............|
              ~~~~~~~~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~ ~~ ~~~~~~~~~~
                HI_MAGIC    zero padding (used to  ↑↑  int('7')
                             be a field here)      ||
                                                   ||
                           ·~~~~~~~~~~~~~~~~~~~~~~~||
                           | version string length ||
                           ·~~~~~~~~~~~~~~~~~~~~~~~~~

    00000010  37 00 00 00 31 00 00 00  30 00 00 00 33 00 00 00  |7...1...0...3...|
            ~~~~ ~~~~~~~~~~~ ~~~~~~~~~~~~ ~~~~~~~~~~~
                  int('1')     int('0')    int('3')


    So the version of this file has 4 characters, and it's 7103. Note how all
    this information is stored as big endian.
    """

    DESCRIPTION = "GHC Haskell .hi files"
    RE_FILE_EXTENSION = re.compile(r"\.(p_|dyn_)?hi$")

    @classmethod
    def recognizes(cls, file):
        if not HiFile.RE_FILE_EXTENSION.search(file.name):
            return False

        if not hasattr(HiFile, "hi_version"):
            try:
                with profile("command", "ghc"):
                    output = our_check_output(["ghc", "--numeric-version"])
            except (OSError, subprocess.CalledProcessError):
                HiFile.hi_version = None
                logger.debug("Unable to read GHC version")
            else:
                major, minor, patch = [
                    int(x) for x in output.decode("utf-8").strip().split(".")
                ]
                HiFile.hi_version = "%d%02d%d" % (major, minor, patch)
                logger.debug("Found .hi version %s", HiFile.hi_version)

        if HiFile.hi_version is None:
            return False

        with open(file.path, "rb") as fp:
            # Read magic
            buf = fp.read(4)
            if buf != HI_MAGIC:
                logger.debug(
                    "Haskell interface magic mismatch. "
                    "Found %r instead of %r or %r",
                    buf,
                    HI_MAGIC_32,
                    HI_MAGIC_64,
                )
                return False

            # Skip some old descriptor thingy that has varying size
            if buf == HI_MAGIC_32:
                fp.read(4)
            elif buf == HI_MAGIC_64:
                fp.read(8)

            # Read version, which is [Char]
            buf = fp.read(1)

            # Small list optimisation - anything less than 0xff has its length
            # in a single byte; everything else is 0xff followed by the 32-bit
            # length (big-endian).
            if buf[0] == 0xFF:
                buf = fp.read(4)
                length = struct.unpack(">I", buf)[0]
            else:
                length = buf[0]

            # Now read characters; each is 32-bit big-endian.
            version_found = "".join(
                chr(struct.unpack(">I", fp.read(4))[0]) for _ in range(length)
            )

            if version_found != HiFile.hi_version:
                logger.debug(
                    "Haskell version mismatch; found %s instead of %s.",
                    version_found,
                    HiFile.hi_version,
                )
                return False

            return True

    def compare_details(self, other, source=None):
        return [Difference.from_operation(ShowIface, self.path, other.path)]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import subprocess

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command, our_check_output
from .utils.libarchive import LibarchiveContainerWithFilelist


@tool_required("isoinfo")
def get_iso9660_names(path):
    return (
        our_check_output(
            (
                "isoinfo",
                "-R",  # Always use RockRidge for names
                "-f",
                "-i",
                path,
            ),
        )
        .strip()
        .split("\n")
    )


class ISO9660PVD(Command):
    @tool_required("isoinfo")
    def cmdline(self):
        return ["isoinfo", "-d", "-i", self.path]


class ISO9660Listing(Command):
    def __init__(self, path, extension=None, *args, **kwargs):
        self._extension = extension
        super().__init__(path, *args, **kwargs)

    @tool_required("isoinfo")
    def cmdline(self):
        cmd = ["isoinfo", "-l", "-i", self.path]

        if self._extension == "joliet":
            cmd.extend(["-J", "-j", "iso8859-15"])
        elif self._extension == "rockridge":
            cmd.extend(["-R"])

        return cmd

    def filter(self, line):
        if self._extension == "joliet":
            return line.decode("iso-8859-15").encode("utf-8")
        return line


class Iso9660File(File):
    DESCRIPTION = "ISO 9660 CD images"
    CONTAINER_CLASSES = [LibarchiveContainerWithFilelist]
    FILE_TYPE_RE = re.compile(r"\bISO 9660\b")

    @classmethod
    def recognizes(cls, file):
        if file.magic_file_type and cls.FILE_TYPE_RE.search(
            file.magic_file_type
        ):
            return True

        # Sometimes CDs put things like MBRs at the front which is an expected
        # part of the ISO9660 standard, but file(1)/libmagic doesn't detect
        # this. <https://en.wikipedia.org/wiki/ISO_9660#Specifications>.
        with open(file.path, "rb") as f:
            f.seek(32769)
            return f.read(5) == b"CD001"

        return False

    def compare_details(self, other, source=None):
        differences = []

        for klass in (ISO9660PVD, ISO9660Listing):
            differences.append(
                Difference.from_operation(klass, self.path, other.path)
            )

        for x in ("joliet", "rockridge"):
            try:
                differences.append(
                    Difference.from_operation(
                        ISO9660Listing,
                        self.path,
                        other.path,
                        operation_args=(x,),
                    )
                )
            except subprocess.CalledProcessError:
                # Probably no joliet or rockridge data
                pass

        return differences



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import collections
import functools
import platform
import shutil

try:
    import distro
except ImportError:
    distro = None

from .profiling import profile
from .external_tools import EXTERNAL_TOOLS, REMAPPED_TOOL_NAMES, GNU_TOOL_NAMES

# The output of --help and --list-tools will use the order of this dict.
# Please keep it alphabetized.
OS_NAMES = collections.OrderedDict(
    [
        ("arch", "Arch Linux"),
        ("debian", "Debian"),
        ("FreeBSD", "FreeBSD"),
        ("guix", "GNU Guix"),
    ]
)


@functools.lru_cache()
def find_executable(cmd):
    """
    Given a command name (eg. `dumppdf`), return the absolute path to that
    command. Will also try the command with some common suffixes (eg.
    `dumppdf.py`) to support distributions that strip or retain them.

    Returns `None` if no command is found.
    """

    for suffix in ("", ".py"):
        val = shutil.which(f"{cmd}{suffix}")
        if val:
            return val


def get_tools(only_missing=False):
    """Return the tool configuration in a dict"""

    d = {}

    external_tools = sorted(tool_required.all)
    if only_missing:
        external_tools = [
            tool for tool in external_tools if not tool_check_installed(tool)
        ]
    d["External-Tools-Required"] = tuple(external_tools)

    current_os = get_current_os()
    os_list = [current_os] if (current_os in OS_NAMES) else iter(OS_NAMES)
    for os_ in os_list:
        tools = set()
        for x in external_tools:
            try:
                tools.add(EXTERNAL_TOOLS[x][os_])
            except KeyError:
                pass
        d["Available-in-{}-packages".format(OS_NAMES[os_])] = tuple(
            sorted(tools)
        )

    d["Missing-Python-Modules"] = tuple(sorted(python_module_missing.modules))

    return d


def get_tool_name(tool):
    return REMAPPED_TOOL_NAMES.get(tool, tool)


def tool_prepend_prefix(prefix, *tools):
    if not prefix:
        return
    for tool in tools:
        REMAPPED_TOOL_NAMES[tool] = prefix + tool


def tool_check_installed(command):
    if (
        command == get_tool_name(command)
        and not os_is_gnu()
        and tool_is_gnu(command)
    ):
        # try "g" + command for each tool, if we're on a non-GNU system
        if find_executable("g" + command):
            tool_prepend_prefix("g", command)

    return find_executable(get_tool_name(command))


def tool_required(command):
    """
    Decorator that checks if the specified tool is installed
    """
    from .exc import RequiredToolNotFound

    if not hasattr(tool_required, "all"):
        tool_required.all = set()
    tool_required.all.add(command)

    def wrapper(fn):
        @functools.wraps(fn)
        def tool_check(*args, **kwargs):
            """
            Due to the way decorators are executed at import-time we defer the
            execution of `find_executable` until we actually run the decorated
            function (instead of prematurely returning a different version of
            `tool_check`).

            This ensures that any os.environ['PATH'] modifications are
            performed prior to the `find_executable` tests.
            """
            if not tool_check_installed(command):
                raise RequiredToolNotFound(command)

            with profile("command", command):
                return fn(*args, **kwargs)

        return tool_check

    return wrapper


def tool_is_gnu(command):
    return command in GNU_TOOL_NAMES


def os_is_gnu():
    system = platform.system()
    return system in ("Linux", "GNU") or system.startswith("GNU/")


def get_current_os():
    system = platform.system()
    if system == "Linux":
        if distro:
            return distro.id()
    return system  # noqa


def get_current_distro_like():
    if not distro:
        return []
    return distro.like().split()


def get_package_provider(tool, os=None):
    try:
        providers = EXTERNAL_TOOLS[tool]
    except KeyError:  # noqa
        return None

    try:
        return providers[get_current_os()]
    except KeyError:
        # lookup failed, try to look for a package for a similar distro
        for d in get_current_distro_like():
            try:
                return providers[d]
            except KeyError:
                pass

    return None


def python_module_missing(name):
    python_module_missing.modules.add(name)


python_module_missing.modules = set()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import abc
import logging

logger = logging.getLogger(__name__)


class Operation(metaclass=abc.ABCMeta):
    def __init__(self, path):
        self.path = path
        self.returncode = None
        self.error_string = None

    @abc.abstractmethod
    def start(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def name(self):
        """
        Name of the operation, e.g. "readelf" or "objdump"
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def full_name(self, truncate=None):
        """
        Full name of the operation, that will be used to call
        command_excluded, e.g. "readelf --debug-dump=info"
        """
        raise NotImplementedError()

    def filter(self, line):
        # Assume operation output is utf-8 by default
        return line

    def terminate(self):
        pass

    def should_show_error(self):
        """
        Whether or not the operation resulted in an unexpected error which
        should be displayed rather than output
        """
        return False

    @abc.abstractproperty
    def output(self):
        raise NotImplementedError()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import os.path
import logging

try:
    from debian.deb822 import Dsc, Deb822
except:
    # Import can fail due to apt_pkg.Error (or similar) tracebacks, but let us
    # re-raise them as ImportErrors so that the Comparator module class
    # handling just works. (See reproducible-builds/diffoscope#300)
    raise ImportError

from diffoscope.changes import Changes
from diffoscope.changes import ChangesFileException
from diffoscope.difference import Difference

from .utils.file import File
from .utils.container import Container

logger = logging.getLogger(__name__)


class DebControlMember(File):
    def __init__(self, container, member_name):
        super().__init__(container)
        self._name = member_name
        self._path = None

    @property
    def container(self):
        return self._container

    @property
    def name(self):
        return self._name

    @property
    def path(self):
        return os.path.join(
            os.path.dirname(self.container.source.path), self.name
        )

    def is_directory(self):
        return False

    def is_symlink(self):
        return False

    def is_device(self):
        return False


class DebControlContainer(Container):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._version_re = DebControlContainer.get_version_trimming_re(self)

    @staticmethod
    def get_version_trimming_re(dcc):
        version = dcc.source._deb822.get("Version")

        # Remove the epoch as it's not in the filename
        version = re.sub(r"^\d+:", "", version)
        if "-" in version:
            upstream, revision = version.rsplit("-", 1)

            return re.compile(
                r"_%s(?:-%s)?" % (re.escape(upstream), re.escape(revision))
            )

        return re.compile(re.escape(version))

    def get_adjusted_members(self):
        for name in self.get_member_names():
            yield self._trim_version_number(name), self.get_member(name)

    def get_member_names(self):
        field = self.source._deb822.get("Files") or self.source._deb822.get(
            "Checksums-Sha256"
        )

        parent_dir = os.path.dirname(self.source.path)

        # Show results from debugging packages last; they are rather verbose,
        # masking other more interesting differences due to truncating the
        # output.
        for name in sorted(
            (x["name"] for x in field),
            key=lambda x: (x.endswith(".deb") and "-dbgsym_" in x, x),
        ):
            # The referenced files are supplied by the user, but the Deb822
            # parser rejects malicious imput such as "../../etc/shadow" etc.
            if not os.path.exists(os.path.join(parent_dir, name)):
                logger.debug(
                    "Could not find file %s referenced in %s",
                    name,
                    self.source.name,
                )
                break

            yield name

    def get_member(self, member_name):
        return DebControlMember(self, member_name)

    def _trim_version_number(self, name):
        return self._version_re.sub("", name)


class DebControlFile(File):
    CONTAINER_CLASSES = [DebControlContainer]

    @staticmethod
    def _get_deb822(file):
        # Be nice to .changes and .dsc comparison in the MissingFile case

        if isinstance(file, DebControlFile):
            return file._deb822

        class NullChanges(dict):
            def get_as_string(self, _):
                return ""

        return NullChanges(Files=[], Version="")

    @staticmethod
    def _parse_gpg(file):
        try:
            with open(file.path) as f:
                header, _, footer = Deb822.split_gpg_and_payload(f)
        except EOFError:
            header = []
            footer = []

        return [b"\n".join(x).decode("utf-8") for x in (header, footer)]

    def compare_details(self, other, source=None):
        gpg_a = self._parse_gpg(self)
        gpg_b = self._parse_gpg(other)

        differences = [
            Difference.from_text(
                gpg_a[0], gpg_b[0], self.path, other.path, source="GPG header"
            )
        ]

        other_deb822 = self._get_deb822(other)

        for field in sorted(
            set(self._deb822.keys()).union(set(other_deb822.keys()))
        ):
            if field.startswith("Checksums-") or field == "Files":
                continue

            my_value = ""
            if field in self._deb822:
                my_value = self._deb822.get_as_string(field).lstrip()

            other_value = ""
            if field in other_deb822:
                other_value = other_deb822.get_as_string(field).lstrip()

            comment = None
            if my_value != other_value and sorted(my_value.split()) == sorted(
                other_value.split()
            ):
                comment = "Ordering differences only"

            differences.append(
                Difference.from_text(
                    my_value,
                    other_value,
                    self.path,
                    other.path,
                    source=field,
                    comment=comment,
                )
            )

        # Compare Files as string
        if self._deb822.get("Files"):
            differences.append(
                Difference.from_text(
                    self._deb822.get_as_string("Files"),
                    other_deb822.get_as_string("Files"),
                    self.path,
                    other.path,
                    source="Files",
                )
            )
        else:
            differences.append(
                Difference.from_text(
                    self._deb822.get_as_string("Checksums-Sha256"),
                    other_deb822.get_as_string("Checksums-Sha256"),
                    self.path,
                    other.path,
                    source="Checksums-Sha256",
                )
            )

        differences.append(
            Difference.from_text(
                gpg_a[1],
                gpg_b[1],
                self.path,
                other.path,
                source="GPG signature",
            )
        )

        return differences


class DotChangesFile(DebControlFile):
    DESCRIPTION = "Debian .changes files"
    FILE_EXTENSION_SUFFIX = {".changes"}

    # .changes files can be identified "data" if they contain non-printable
    # characters (Re: reproducible-builds/diffoscope#286)
    #
    # In addition, different versions of file(1) can emit "UTF-8 Unicode text"
    # or "UTF-8 Unicode text" for .changes files that use non-ASCII characters,
    # so we have to look for both. (Re: reproducible-builds/diffoscope#291)
    FILE_TYPE_RE = re.compile(
        r"^(ASCII text|UTF-8 Unicode text|data|Unicode text, UTF-8 text)"
    )

    @classmethod
    def recognizes(cls, file):
        if not super().recognizes(file):
            return False

        try:
            file._deb822 = Changes(filename=file.path)
        except ChangesFileException as exc:
            logger.warning(
                f"Rejecting {file.path} as a Debian .changes file: {exc}"
            )
            return False

        try:
            file.validation_msg = None
            file._deb822.validate("sha256", check_signature=False)
        except FileNotFoundError as exc:
            # Continue even though this .changes file may be invalid
            file.validation_msg = f"{os.path.basename(file.path)} is missing referenced files: {exc}"
            logger.warning(file.validation_msg)

        return True

    def compare(self, other, *args, **kwargs):
        differences = super().compare(other, *args, **kwargs)

        if differences is None:
            return None

        for x in (self, other):
            if x.validation_msg:
                differences.add_comment(x.validation_msg)

        other_deb822 = self._get_deb822(other)

        files = zip(self._deb822.get("Files"), other_deb822.get("Files"))

        files_identical = all(
            x == y for x, y in files if not x["name"].endswith(".buildinfo")
        )

        if (
            files_identical
            and len(differences.details) == 1
            and differences.details[0].source1 == "Files"
        ):
            logger.warning("Ignoring buildinfo file differences")
            return None

        return differences


class DotDscFile(DebControlFile):
    DESCRIPTION = "Debian source packages (.dsc)"
    FILE_EXTENSION_SUFFIX = {".dsc"}

    @classmethod
    def recognizes(cls, file):
        if not super().recognizes(file):
            return False

        with open(file.path, "rb") as f:
            file._deb822 = Dsc(f)

        return True


class DotBuildinfoContainer(DebControlContainer):
    def get_member_names(self):
        result = super(DotBuildinfoContainer, self).get_member_names()

        # As a special-case, if the parent container of this .buildinfo is a
        # .changes file, ignore members here that are referenced in both. This
        # avoids recursing into files twice where a .buildinfo references a
        # file that is also listed in that member's parent .changes file:
        #
        #    foo.changes → foo.deb
        #    foo.changes → foo.buildinfo → foo.deb
        #
        ignore = set()
        if isinstance(self.source.container, DebControlContainer):
            ignore.update(self.source.container.get_member_names())

        return [x for x in result if x not in ignore]


class DotBuildinfoFile(DebControlFile):
    DESCRIPTION = "Debian .buildinfo files"
    CONTAINER_CLASSES = [DotBuildinfoContainer]
    FILE_EXTENSION_SUFFIX = {".buildinfo"}
    FILE_TYPE_RE = re.compile(
        r"^(ASCII text|UTF-8 Unicode text|Unicode text, UTF-8 text|PGP signed message)"
    )

    @classmethod
    def recognizes(cls, file):
        if not super().recognizes(file):
            return False

        # Parse .buildinfo files like .dsc files
        with open(file.path, "rb") as f:
            file._deb822 = Dsc(f)

        # Check we have a *Debian* .buildinfo file
        if not (
            "Format" in file._deb822 and "Checksums-Sha256" in file._deb822
        ):
            return False

        return True
